/*
checkdates: Birthday reminder program
Michael Ernst <mernst@alum.mit.edu> 8/9/90

When compiling, you will want to change the (hardcoded, alas) constant
"addressfile".

This program reads a file containing date information and lists all dates 
within a specified number of days before or after today.
For usage information, use "-h" flag or see exit_print_usage below.

One potential use is to put a line like the following in your crontab entry:
  26 4 * * *	checkdates | mail -s "Checkdates" myaddress@myhost.com
This causes you to be sent a reminder of relevant dates every day at 4:26am.

The input file is composed of paragraphs separated by blank lines.
The first line of each paragraph is an individual's name, last name first,
or the names of a couple separated by &, in which case the second name may
be followed by "(nee MAIDENNAME)" or "(no nee)", the latter indicating
differing last names.  For instance,
    Ball, Tom
    Birkhimer, Neil & Lisa
    Benco, Michael `Ouija' & Andrea (nee Born)
    Friedman, Julio & Lara Webber (no nee)
The second line may list children, names separated by commas, each optionally
followed by (now MARRIEDNAME).
    Karl, Ellen (now Brenden), Neil, Nancy
The remainder of the paragraph (including the second line, which is
interpreted as children only if necessary) is free-form.  Cross-referencing
name changes is useful because it prevents multiple mention of a single
individual.

The paragraph is ignored unless one of the lines begins with "dates: ".  If
such a line exists, it is a semicolon-separated list of dates.  Parenthesized
dates are ignored (and are not even parsed, so need not be valid dates),
which is helpful for storing dates you don't want to be reminded of; 
non-parenthesized dates are interpreted in a relatively flexible way.
For instance, all of the following are valid:
  11/8/39	5/7	6/30/1968	16 Mar 1963	July 19, 1963
Additional parenthetical material may follow any date and is ignored:
  1/29 (around 1960)
The dates are interpreted as: husband's birthday, wife's birthday,
anniversary, children's birthdays.  Dates may be empty (ie, may be
blank), and as many or as few as desired may be supplied.  For instance:
    dates: 7/7
    dates: (6/23)
    dates: 9/28/68 (8:26 AM, New York)
    dates: 10/7/59; ; 6/30/96; (1/20/90); (4/21/92) */

/*
  I should just punt this in favor of something in gawk or elisp or something.
  */

/* I should change the source from addresses.tex to just addresses.
   Then change printaddresses, too.
*/

#define DEBUG 0

#include <stdio.h>
#include <time.h>
/* #include <string.h> */		/* extraneous */
#include <errno.h>
#include "c-type.h"
#include "c-util.h"
#include "c-util.c"
#include "dateutil.c"

/* global variables */

#define STRINGMAX	(100)
#define DATESSTRINGMAX	(150)

int days_before = 4;
int days_after = 10;

int today_yeardate;

char *addressfile = "/homes/gws/mernst/random/addresses.tex";
/* char *addressfile = "/home/mernst/random/addresses.tex"; */
/* char *addressfile = "c:\\random\\address.tex"; */

char line1[STRINGMAX] = "";
char line2[STRINGMAX] = "";

Bool dont_ignore = FALSE;
Bool allyear = FALSE;
Bool texprint = FALSE;

/* A huge amount of memory, ouch! */
char dates[366][DATESSTRINGMAX];

/* The default should be that non-flag command-line arguments are treated
   as input or output file??
   If that becomes the case, then perhaps I should standardize the flag usage.
*/

/* I should prevent the extra \\ from being printed 
   before the first month name
*/

void exit_print_usage()
{ 
  fprintf(stderr,"Usage:  checkdates [-+<digits>ypifoh]\n\n");
  fprintf(stderr,"  Prints all dates within a specified number of days before or after today.\n");
  fprintf(stderr,"  Dates out of range or in parentheses are ordinarily ignored.\n");
  fprintf(stderr,"Flag usage:\n");
  fprintf(stderr," -        next number means days before\n");
  fprintf(stderr," +        next number means days after\n");
  fprintf(stderr," <digits> interpreted as a number, days before or after\n");
  fprintf(stderr," y        all dates, irrespective of range (full year checking)\n");
  fprintf(stderr," p        write output to a TeX file; this implies all dates\n");
  fprintf(stderr," i        don't ignore parenthesized dates\n");
  fprintf(stderr," f        use next argument as input filename (not implemented)\n");
  fprintf(stderr," o        use next argument as output filename (not implemented)\n");
  fprintf(stderr," h        print a help message, then exit\n");
  exit(1);
}


/* Takes the index of the name, returns a string. */
char *getname(int dateno)
{
  /* Static so we can reuse it every time without worrying about freeing it. */
  static char result[STRINGMAX];
  char lastname[STRINGMAX];
  char *husband, *wife, *tmp;
  char child[STRINGMAX];
  int nonee = FALSE;

  /* Don't clobber line1 */
  strcpy(lastname, line1);
  /* Fails if no first name (eg, Valentine's Day) */
  strtok(lastname, ",");
  strwrtrim(lastname);
  husband = strtok(NULL, "&");
  if (husband == NULL)
    {
      husband = "";
      wife = "";
    }
  else
    {
      husband = strwtrim(husband);
      /* just to return the internal state pointer */
      wife = strtok(NULL, "\0");
      /* strstr is supposed to work even if first arg is NULL, but it fails. */
      if (wife != NULL)
	{
	  if ((tmp = strstr(wife, " (nee")))
	    tmp[0] = '\0';
	  if ((tmp = strstr(wife, " (no nee)")))
	    {
	      nonee = TRUE;
	      tmp[0] = '\0';
	    }
	}
    }
  
  if (DEBUG)
    printf("getname: `%s' `%s' `%s' %d\n", lastname, husband, wife, nonee);

  switch (dateno) {
    case 1:
      /* Husband */
      if ((husband != NULL) && (*husband != '\0'))
	sprintf(result, "%s %s", husband, lastname);
      else /* perhaps a holiday */
	sprintf(result, "%s", lastname);
      break;
    case 2:
      /* Wife */
      if (nonee)
	sprintf(result, "%s", strwtrim(wife));
      else
	sprintf(result, "%s %s", strwtrim(wife), lastname);
      break;
    case 3:
      /* Anniversary */
      if (wife == NULL)
	{ sprintf(result, "Warning: no wife for %s %s", husband, lastname);
	  wife = ""; }
      if (nonee)
	sprintf(result, "%s %s & %s", husband, lastname, strwtrim(wife));
      else
	sprintf(result, "%s & %s %s", husband, strwtrim(wife), lastname);
      break;
    default:
      /* Child */
      strcpy(child, line2);
      tmp = strtok(child, ",");
      while (dateno-- > 4)
      {
	tmp = strtok(NULL, ",");
      }
      strcpy(child, tmp);
      if ((tmp = strstr(child, "(now ")))
	{
	  strcpy(lastname, tmp+5);
	  strwrtrim(lastname);		/* ( editor */
	  if (lastname[strlen(lastname)-1]==')')
	    lastname[strlen(lastname)-1] = '\0';
	  tmp[0] = '\0';
	}
      sprintf(result, "%s %s", strwtrim(child), lastname);
      break;
  }

  return result;
}


/* the "dates:" has already been removed from the string */
void processdates(char *datesline, int today_yeardate)
{
  int month, day, year;
  int yeardate;
  char formatted_name[STRINGMAX];
  int dateno = 0;
  char *semipos;

  /* Assume there's at least one date */
  do {
    if ((semipos = strchr(datesline, ';')) != NULL)
      semipos[0] = '\0';
    dateno++;
    datesline = strwtrim(datesline);
    if (*datesline == '(')
      if (dont_ignore)
	/* we're not ignoring parenthesized dates; remove the parentheses */
	{
	  datesline++;
          /* this is necessary, lest "Oct)" is seen as a month */
	  if (datesline[strlen(datesline)-1] == ')')
	    datesline[strlen(datesline)-1] = '\0';
	  else 
	    fprintf(stderr, "Last character should have been parenthesis, was `%c'=%d: (%s",
		    /* matching ) for the editor */
		    datesline[strlen(datesline)-1], datesline[strlen(datesline)-1], datesline);
	}
      else
	/* else we're ignoring parenthesized dates; null out the string */
	datesline[0] = '\0';
      
    if (DEBUG)
      printf("processdates date: `%s' (length %d)\n", datesline, strlen(datesline));

    if (!strempty(datesline))
      if (!parsedate(datesline, &month, &day, &year))
	fprintf(stderr, "Couldn't parse date `%s' associated with %s\n",
		datesline, line1);
      else
	{
	  if (DEBUG)
	    printf("parsedate rturned; m=%d, d=%d, y=%d\n", month, day, year);
	  
	  yeardate = day_of_year(month, day);
	    
	  if (allyear
	      || ((366 + today_yeardate - yeardate)%366 <= days_before)
	      || ((366 + yeardate - today_yeardate)%366 <= days_after))
	  { 
	    sprintf(formatted_name, date_to_string(month, day, year),
		    getname(dateno));
	    yeardate = yeardate % 366;

	    /* if this is identically there, don't add it */
	    if (((strempty(dates[yeardate]))
		 || (NULL == strpfind(dates[yeardate], formatted_name)))
		&& (strlen(dates[yeardate]) + strlen(formatted_name) < DATESSTRINGMAX))
	      strcat(dates[yeardate], formatted_name);
	  }
	}
    /* This is illegal if semipos == NULL, but hen we drop out of the loop */
    datesline = semipos+1;

  } while (semipos != NULL);


  /* done with dates on this line */
}

void readlines()
{
  char thisline[STRINGMAX];
  char *trimmedline;

  FILE *infile;

  infile = fopen(addressfile,"r");

  if (infile == NULL)
    { fprintf(stderr, "Problem reading file %s; error code %d.\n",
	      addressfile, errno);
      exit(1);
    }

  while (NULL != fgets(thisline, STRINGMAX, infile))
    {
      trimmedline = strwltrim(thisline);
      
      if (strempty(trimmedline))
	/* new paragraph; zero out line1 and line2 */
	{ line1[0] = '\0';
	  line2[0] = '\0'; }
      else if (strncmp(trimmedline, "dates:", 6) == 0)
	{ trimmedline = trimmedline + 6;
	  processdates(trimmedline, today_yeardate);
	}
      else if (strempty(line1))
	strcpy(line1, trimmedline);
      else if (strempty(line2))
	strcpy(line2, trimmedline);
    }
  fclose(infile);
}

void initialize(int argc,char *argv[])
{
  { time_t clock;
    struct tm *tm;
    
    /* get the real time into tm */
    clock = (time_t) time(0);
    tm = localtime(&clock);
    /* +1 because I use base 1 for month indexing */
    today_yeardate = day_of_year((tm->tm_mon)+1, tm->tm_mday);
    
    datesetstyle_internal("|n | D %s |'Y\n");
  }

  { int i;
    for (i=0; i<366; i++)
      dates[i][0] = '\0';
  }

  { /* get command line options */
    Bool setafter = FALSE;
    char c;
    
    /* argc = # of arguments, +1 (for the program name itself) */
    while (--argc > 0)
      { argv++;
	while ((c = *(argv[0]++)))
	  if isdigit(c)
	    { int value = c - '0';
	      while (isdigit(c = *(argv[0]++)))
		value = 10 * value + c - '0';
	      argv[0]--;	/* 'ungetc'-like action */
	      /* printf(" %d\n", value); */
	      if (setafter)
		days_after = value;
	      else
		days_before = value;
	    }
	  else
	    switch (c) {
	    case '-':
	      setafter = FALSE;
	      break;
	    case '+':
	      setafter = TRUE;
	      break;
	    case 'i':
	      dont_ignore = TRUE;
	      break;
	    case 'p':
	      texprint = TRUE;
	      /* note fallthrough */
	    case 'y':
	      allyear = TRUE;
	      break;
	    default:   /* -h or unknown option, let's quit */
	      fprintf(stderr,"Unrecognized option %c.\n",c);
	    case 'h':
	      exit_print_usage();
	    }
      }    
  }
}

/* ACTION:  have an option, or perhaps always, show month only if this
   starts a new one */
/* I should quite possibly be generating the date here rather than before */

void printdates()
{
  int i;

  if (allyear)
    for (i=1; i<=366; i++)
      printf("%s", dates[i % 366]);
  else
    for (i=today_yeardate - days_before; i<=today_yeardate + days_after; i++)
      printf("%s", dates[(i+366) % 366]);
}

/* This is a hack */
void texprintdates()
{
  char month[4] = "   \0";

  int i;

  FILE *outfile = fopen("/u/mernst/adates.tex","w");

  if (outfile == NULL)
    { fprintf(stderr, "Problem writing file %s; error code %d.\n",
	      "/u/mernst/adates.tex", errno);
      exit(1);
    }

  /* header of file */
  /* change this to use multicol.sty */
  fprintf(outfile, "%s", "% -*- Mode: TeX -*- \n% for LaTeX\n\\documentstyle[twocolumn]{article}\n\n\\input text-macros\n\\mypagesize\n\\pagestyle{empty}\n\\catcode`\\&=12\n\n\\begin{document}\n\n\\begin{tabbing}\nMay \\= 22 \\= \\kill\n");

  for (i=1; i<=366; i++)
    { char *temp = strtok(dates[i % 366], "\n");

      while (temp != NULL)
	{
	  if (strncmp(month, temp, 3))
	    { 
	      strncpy(month, temp, 3);
	      fprintf(outfile, " \\\\\n%s", month);
	    }
	  else
	    fprintf(outfile, "   ");

	  temp = temp + 4;

	  /* This may be gambling with execution order */
	  fprintf(outfile, " \\> %c%c", *temp++, *temp++);
	  fprintf(outfile, " \\>%s \\\\\n", temp);

	  temp = strtok(NULL, "\n");
	}
    }

  /* tail of file */
  fprintf(outfile, "\\end{tabbing}\n\\end{document}\n");

  fclose(outfile);
}





void main(int argc, char *argv[])
{
  initialize(argc, argv);

  readlines();

  if (texprint)
    { texprintdates();
      /* I'd like to call TeX here, too, but I'm not sure how to do it */
    }
  else
    printdates();
}

/* For GNU Emacs */
/* Local Variables: */
/* compile-command: "gcc -g -O -Wall checkdat.c -o checkdat-new -I/homes/gws/mernst/bin/src/util -lm" */
/* compile-history: ("gcc checkdat.c -o checkdat-new -I/homes/gws/mernst/bin/util -lm -g -O -Wall") */
/* End: */
