/* 
   Date conversion utilities
   Michael D. Ernst, mernst@theory.lcs.mit.edu
*/


#include <stdio.h>

/* This describes a leap year, which shouldn't matter.  1-based, not
0-based. */
static int days_preceeding[13] =
  /*  {0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31}; */
  /*  {0, 0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334}; */
  {0, 0, 31, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335};

int day_of_year(int month, int day)
{  
  return(day + days_preceeding[month]);
}

void month_and_day(int yearday, int *month, int *day)
{
  /* *month-- and *month++ are wrong because * has lower precedence */
  for (*month=1; *month<13; (*month)++)
    if (days_preceeding[*month] >= yearday)
      { (*month)--;
	*day = yearday - days_preceeding[*month];
	return;
      }
  *month = 0;
  *day = 0;
  return;
}

static char *month_abbrev[13] = { "",
  "Jan", "Feb", "Mar", "Apr", "May", "Jun",
  "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };

static char *month_name[13] = { "",
  "January", "February", "March", "April", "May", "June",
  "July", "August", "September", "October", "November", "December" };

/* I could check that the day is reasonable here */
/* Returns TRUE if a date was found, FALSE otherwise */
/* Returns the year as is; does not change 76 to 1976, 1776, or anything else */
/* ACTION:  deal with st, nd, rd, th */

/* ACTION:  completely rewrite this to deal with one word at a time.
   Make it work like the lispm's.
   I probably won't ever do this, as what I have is enough for my needs.
*/
Bool parsedate(char *date, int *monthptr, int *dayptr, int *yearptr)
{
  char monthname[20];

  if (strempty(date))
    return FALSE;

  if (DEBUG) printf("parsedate `%s'\n", date);
  *yearptr = 0;
  if (sscanf(date, "%d/%d/%d", monthptr, dayptr, yearptr) == 3)
    return TRUE;
  if (sscanf(date, "%d/%d", monthptr, dayptr) == 2)
    return TRUE;
  /* ACTION:  I should deal with 1st, 2nd, 3rd, 4th here. */
  if ((sscanf(date, "%d %s '%d", dayptr, monthname, yearptr) == 3)
      || (sscanf(date, "%d %s %d", dayptr, monthname, yearptr) == 3)
      || (sscanf(date, "%d %s", dayptr, monthname) == 2)
      || (sscanf(date, "%s %d, '%d", monthname, dayptr, yearptr) == 3)
      || (sscanf(date, "%s %d, %d", monthname, dayptr, yearptr) == 3)
      || (sscanf(date, "%s %d '%d", monthname, dayptr, yearptr) == 3)
      || (sscanf(date, "%s %d %d", monthname, dayptr, yearptr) == 3)
      || (sscanf(date, "%s %d", monthname, dayptr) == 2))
    { 
      /* don't just check for 3-letter month abbreviations but for 
	 4-letter ones (e.g. Sept) as well
      */

      if (monthname[strlen(monthname)-1] == '.')
	monthname[strlen(monthname)-1] = 0;
      for (*monthptr=1; *monthptr<13; (*monthptr)++)
	if (strequal(monthname, month_abbrev[*monthptr])
	    || strequal(monthname, month_name[*monthptr]))
	  return TRUE;
    }
  return FALSE;
}

/* Maybe make this "smart"; that would be nice. */
/* That is, a "print using" sort of style. */
/* Require that the date be February 8, 1968 */
/* Later on, be more lenient, and make assumptions wrt "May", "14", etc. */

char *datestyle;

/* Commented out for the nonce, as it isn't implemented (or used).
   See datestyle_internal, below. */
/*char *datesetstyle(char *style)
{
*/
/* 
  My print using:  
  || the | character itself
     options between | and |:
     <digits>  number of |'s to write
  |N full month name
     options between | and N:
  |n abbreviated (3 character default) month name
     options between | and n:
     .  add a . after the abbreviation
     ,  as . flag, but complete the month name if possible (June not Jun.)
        (ie, space) pad with spaces on the right
     <digits>  number of characters in abbreviation (|N is equivalent to |9n)
  |M month number, no padding
  |D day
  |Y year
     options between | and M, D, or Y:
     0  pad to 2 digits with 0's
        (ie, space) pad to [2,2,4] digits with spaces
     o  append st, nd, th, or st, as appropriate, to make ordinal
     2  set value = value % 100; ie, last two digits only; implies 0 option
     '  add ' if < 100 (or if 2 option is specified); implies 0 option
     ? start omitting if value = 0 (no output directly produced)
     ! end omitting (no output directly produced)

*/
  /* ACTION:  implement this */

  /* perhaps have an option to call this a second time, or with a second
     argument, to clarify ambiguities
  */

  /* call datesetstyle_internal */
/*
}
*/


/* For now I'll use this since I'm too lazy to finish the smart parser */
void datesetstyle_internal(char *style)
{
  datestyle = strdup(style);
}

/* I ought to keep count to make sure I don't go over 200 characters. */
char *date_to_string(int month, int day, int year)
{ 
 char *result = malloc(200);
 char *resultptr = result;
 char *styleptr = datestyle;
 char *tocopy;
 char tocopyarray[10];

 Bool omitting = FALSE;
 Bool omittingmonth = FALSE;
 Bool omittingday = FALSE;
 Bool omittingyear = FALSE;
 Bool in_escape_sequence = TRUE;

 Bool periodflag = FALSE;
 Bool abbrevcompletingflag = FALSE;
 Bool spaceflag = FALSE;
 Bool zeroflag = FALSE;
 Bool twoflag = FALSE;
 Bool digitflag = FALSE;
 Bool ordinalflag = FALSE;
 Bool apostropheflag = FALSE;
 Bool omittingbeginflag = FALSE;
 Bool omittingendflag = FALSE;
 int numericarg = 0;

 int value = 0;	/* the month, day, or year number */

 while (*styleptr)
   {
     if (*styleptr == '|')
       {
	 /* zero out tocopyarray, which is output after the while loop */
	 tocopyarray[0] = '\0';

	 /* this is for debugging only */
	 if (periodflag || abbrevcompletingflag || spaceflag || zeroflag
	     || twoflag || digitflag || ordinalflag || apostropheflag
	     || omittingbeginflag || omittingendflag || numericarg)
	   { error("Something isn't zeroed out.");
	     printf("%d %d %d %d %d %d %d %d %d %d %d\n",
		    periodflag, abbrevcompletingflag, spaceflag, zeroflag,
		    twoflag, digitflag, ordinalflag, apostropheflag,
		    omittingbeginflag, omittingendflag, numericarg);
	     /*
	     spaceflag = FALSE;
	     zeroflag = digitflag = FALSE;
	     */
	   }

	 in_escape_sequence = TRUE;
	 while (in_escape_sequence)
	   
	   switch (*++styleptr)
	     { /* styleptr now points to the character in question */
	     case ',':
	       abbrevcompletingflag = TRUE;
	       /* fall through */
	     case '.':
	       if (periodflag)
		 error("It is not sensible to have 2 periods, 2 commas, or a period and a comma.");
	       else
		 periodflag = TRUE;
	       break;
	     case '0':
	       zeroflag = TRUE;
	       numericarg = numericarg * 10;
	       digitflag = TRUE;
	       break;
	       /* fall through */
	     case ' ':
	       if (spaceflag)
		 error("You should only specify the space flag once.");
	       spaceflag = TRUE;
	       break;
	     case '2':
	       twoflag = TRUE;
	       /* fall through */
	     case '1':
	     case '3':
	     case '4':
	     case '5':
	     case '6':
	     case '7':
	     case '8':
	     case '9':
	       digitflag = TRUE;
	       numericarg = numericarg * 10 + (*styleptr - '0');
	       break;
	     case 'o':
	       ordinalflag = TRUE;
	       break;
	     case '\'':
	       apostropheflag = TRUE;
	       zeroflag = TRUE;
	       break;
	     case '?':
	       omittingbeginflag = TRUE;
	       break;
	     case '!':
	       omittingendflag = TRUE;
	       break;
	       
	     case '|':
	       if (periodflag || spaceflag || ordinalflag || apostropheflag
		   || omittingbeginflag || omittingendflag)
		 { error("Bad flag to ||...");
		   periodflag = spaceflag = ordinalflag = apostropheflag
		     = omittingbeginflag = omittingendflag = FALSE;
		 }

	       /* OR:  set my flags false.  any that are true, are bad. */
	       /*  still check for conflicting ones here */
	       
	       if (!omitting)
		 { if (!digitflag)
		     numericarg = 1;
		   for (; numericarg-- > 0; )
		     *resultptr++ = *styleptr;
		 }
	       in_escape_sequence = FALSE;
	       /* printf("escaping via '|'\n"); */
	       break;
	     case 'N':
	       if (periodflag || spaceflag || digitflag)
		 { error("There should be no modifiers (such as period, "
			 "comma, space, or digit) to |N");
		   periodflag = abbrevcompletingflag = spaceflag = FALSE;
		 }
	       digitflag = TRUE;
	       numericarg = 9;

	       if (value)
		 error("Value!! %d\n",value);
	       /* fall through */
	     case 'n':
	       
	       if (ordinalflag || apostropheflag
		   || omittingbeginflag || omittingendflag)
		 { error("The o, ', ?, and ! flags are not valid for"
			 "escape sequence |%c.", *styleptr);
		   ordinalflag = apostropheflag
		     = omittingbeginflag = omittingendflag = FALSE;
		 }

	       if (!digitflag)
		 numericarg = 3;
	       
	       if (numericarg < 3)
		 error("numeric arg should be >= 3");
	       else
		 if (!omitting)
		   { int goallength
		       /* was (spaceflag && periodflag) */
		       = numericarg + (periodflag ? 1 : 0);
		     int month_name_len = strlen(month_name[month]);
		     
		     memcpy(tocopyarray, month_name[month], goallength);
		     /* There may be an earlier \0 that takes precedence. */
		     tocopyarray[goallength] = '\0';
		     
		     if (month_name_len <= numericarg)
		       /* month name can appear in full; padding may
be required */
		       { /* this set of brackets to keep the else straight */
			 if (spaceflag)
			   { int i;
			     for (i = month_name_len; i < goallength; i++)
			       tocopyarray[i] = ' ';
			   }
		       }
		     else /* (month_name_len > numericarg) */
		       /* name may appear incompletely; no padding */
		       if (periodflag
			   && !(abbrevcompletingflag
				&& (month_name_len == goallength)))
			 tocopyarray[numericarg] = '.';
		   }
	       digitflag = zeroflag = twoflag = spaceflag = FALSE;
	       numericarg = 0;

	       in_escape_sequence = FALSE;
	       /* printf("escaping via 'n' (%c)\n", *styleptr); */
	       break;
	       
	     case 'M':
	       if (omittingbeginflag)
		 if (omittingmonth)
		   error("Shouldn't have successive |?M without a separating |!M.");
		 else if (month == 0)
		   omitting = omittingmonth = TRUE;
	       if (omittingendflag)
		 omittingmonth = FALSE;
	       value = month;
	       /* havevalue = TRUE; */
	       in_escape_sequence = FALSE;
	       /* printf("escaping via 'M' (%c)\n", *styleptr); */
	       break;
	     case 'D':
	       if (omittingbeginflag)
		 if (omittingday)
		   error("Shouldn't have successive |?D without a separating |!D.");
		 else if (day == 0)
		   omitting = omittingday = TRUE;
	       if (omittingendflag)
		 omittingday = FALSE;
	       value = day;
	       /* havevalue = TRUE; */
	       in_escape_sequence = FALSE;
	       /* printf("escaping via 'D'\n"); */
	       break;
	     case 'Y':
	       if (omittingbeginflag)
		 if (omittingyear)
		   error("Shouldn't have successive |?Y without a separating |!Y.");
		 else if (year == 0)
		   omitting = omittingyear = TRUE;
	       if (omittingendflag)
		 omittingyear = FALSE;
	       value = year;
	       /* havevalue = TRUE; */
	       in_escape_sequence = FALSE;
	       /* printf("escaping via 'Y'\n"); */
	       break;
	       
	     default:
	       error("Unknown indicator %c after | being ignored.", *styleptr);
	       /* actually, here (and elsewhere) I should print out the
		  entire format string and  show where the difficulty is
	       */
	       periodflag = abbrevcompletingflag = spaceflag = zeroflag
		 = twoflag = digitflag = ordinalflag = apostropheflag
		   = omittingbeginflag = omittingendflag = FALSE;
	       numericarg = 0;

	       in_escape_sequence = FALSE;
	       /* printf("escaping via error\n"); */
	       break;
	     } /* end switch, while */
	 
	 /*
	   omittingflag is true; do omittingeval
	   value is true; have a value to be processed
	   tocopyarray has stuff to be printed;
	   just printed "|";

	   alternately: (I don't do this)
	   havevalue is true; have a value to be processed
	   tocopyarray has stuff to be printed;
	   just printed "|";
	   */
	 
	 if ((omittingbeginflag && omittingendflag) || 
	     ((omittingbeginflag || omittingendflag) 
	      && (periodflag || abbrevcompletingflag || spaceflag || zeroflag
		  || twoflag || digitflag || ordinalflag || apostropheflag)))
	   { error("When ? or ! appears, it must be the only modifier to %c",
		   *styleptr);
	     /* ought to zero them all out here */
	   }

	 if (omittingbeginflag)
	   omittingbeginflag = FALSE;
	 else if (omittingendflag)
	   { omitting = (omittingmonth || omittingday || omittingyear);
	     omittingendflag = FALSE;
	   }
	 else if (value && !omitting)
	   {
	     /* no need to check for comma, since comma => period */
	     if (!strempty(tocopyarray))
	       error("There was something here that got mashed.\n%s\n",
		     tocopyarray);

	     if (periodflag)
	       { error(". and , are not valid modifiers to %c",*styleptr);
	       }
	     if ((numericarg && (numericarg != 2)) || (zeroflag && twoflag))
	       { error("The only valid digit modifiers are 0 and 2, "
		       "at most one of which may be specified.");
		 numericarg = 0;
	       }
	     
	     if (twoflag)
	       value = value % 100;
	     
	     /* This doesn't treat years like 3 the way you might want, when
		it's really the year 3 and not '03.  But I suspect that that
		will rarely or never come up, so it's not a problem.
		*/
	     
	     if (apostropheflag)
	       { if (value < 100)
		   { if (spaceflag)
		       *resultptr++ = ' ';
		     *resultptr++ = '\'';
		   }
	       }
	     
	     if (value < 10)
	       if (zeroflag || twoflag)
		 *resultptr++ = '0';
	       else if (spaceflag)
		 *resultptr++ = ' ';
	     
	     sprintf(tocopyarray, "%d",value);
	     
	     if (ordinalflag)
	       { switch (value % 100)
		   {
		   case 11:
		   case 12:
		   case 13:
		     strcat(tocopyarray, "th"); break;
		   default:
		     switch (value % 10)
		       {
		       case 1:  strcat(tocopyarray, "st"); break;
		       case 2:  strcat(tocopyarray, "nd"); break;
		       case 3:  strcat(tocopyarray, "rd"); break;
		       default: strcat(tocopyarray, "th"); break;
		       }
		   }
	       }
	   } /* end if (value && !omitting) */
	 /* clear out all the flags that might still be set */
	 periodflag = abbrevcompletingflag = FALSE;
	 zeroflag = twoflag = digitflag = spaceflag = FALSE;
	 numericarg = 0;
	 apostropheflag = FALSE;
	 ordinalflag = FALSE;
	 value = 0;
	 
	 /* output tocopy */
	 
	 /*
	   output tocopyarray, if any;
	   */
	 
	 /* no need to make the omitting check; that's already been done */
	 /* where to put it?  if above, less work but slightly bigger code */
	 tocopy = tocopyarray;
	 while ((*resultptr++ = *tocopy++))
	   ;
	 resultptr--;
	 
       }
     else /* (*styleptr =! '|') */
       if (!omitting)
	 *resultptr++ = *styleptr;

     
     styleptr++;
   } /* end of while loop, end of style string */

 *resultptr = '\0';
 return result;
}

